%module compass

%inline %{
/* compass.i */
%}

%{
/* For Numeric. */
#define PY_ARRAY_UNIQUE_SYMBOL compass_UNIQUE_SYMBOL
#include "Python.h"
#include "Numeric/arrayobject.h"
%}

%include typemaps.i

%init %{
    import_array();
%}

%include ARR.i
%include defines.i
%include bs.i
%include emd.i
%include convert.i
%include outputarrs.i

%{
/* Modified by Edward C. Jones, April, 2002 */

/***************************************************************************
 * The Compass Operator -- an algorithm to find edges and corners in color *
 *                   images by representing regions with distributions     *
 *                                                                         *
 * Mark A. Ruzon                                                           *
 * Stanford Vision Laboratory                                              *
 * October, 1997                                                           *
 * Copyright 1997-1999.  All rights reserved.                              *
 *                                                                         *
 * Details -- The compass operator uses a circular window centered at a    *
 *   junction where 4 pixel squares meet.  The needle is a diameter at a   *
 *   given orientation.  The color distributions of the two semicircles    *
 *   are computed, and the distance between them is measured using the     *
 *   Earth Mover's Distance (EMD).  The maximum EMD over all orientations  *
 *   gives the edge strength and orientation at that point.  Uncertainty   *
 *   and abnormality are also measured.                                    *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "compass.h"

#define MIN(x,y) ((x) < (y)) ? (x) : (y)
#define PI 3.14159265358979323846
#define MAXRADIUS 64
#define PERCEPTUAL_THRESH 11.0
#define GAMMA 14.0
#define BLACK 40
#define SAMPLE_PERCENTAGE 1.0
#define WEDGEWT 1.0
#define DEBUG 0
#define PARTIAL_MATCH 1
#define PRINT_LOC 0
#define PRINT_SAMPLING 0
#define PRINT_CLUSTERS 0
#define PRINT_COST 0
#define PRINT_HIST 0

/* This global is for the EMD cost matrix */
double cost[MAXCLUSTERS][MAXCLUSTERS];

void RGB2Lab(float, float, float, float *, float *, float *);

/**************************************************/
/***     Level 2 Functions                      ***/
/**************************************************/

/* This routine abstracts a little of the complexity of placing pixels and
 * their weights into the unclustered array.
 */
static void
AddWeight(PyArrayObject* Lab,
          double weight, int row, int col,
          PyArrayObject* data, PyArrayObject* hist, int wedge,
          float com[][DIM], float sum[], int nclusters)
{
    int min_index = -1, i;
    double dist, Ldist, adist, bdist, min_dist = 100 * 240 * 240;

    for (i = 0; i < nclusters; i++) {
        Ldist = LabSUB(0, row, col) - dataSUB(i, 0);
        adist = LabSUB(1, row, col) - dataSUB(i, 1);
        bdist = LabSUB(2, row, col) - dataSUB(i, 2);
        dist = Ldist * Ldist + adist * adist + bdist * bdist;
        if (dist < min_dist) {
            min_dist = dist;
            min_index = i;
        }
    }

    histSUB(wedge, min_index) += weight;
/*    ckindex("L110", min_index, MAXCLUSTERS); */
    sum[min_index] += weight;
    com[min_index][0] += weight * LabSUB(0, row, col);
    com[min_index][1] += weight * LabSUB(1, row, col);
    com[min_index][2] += weight * LabSUB(2, row, col);

}

/* Distance function for the EMD */
static float
dist(feature_t *F1, feature_t *F2) 
{ 
    return((float)cost[*F1][*F2]);
}

/* Order statistics function, from Cormen, Leiserson, and Rivest, Ch. 8 & 10 */
/* ecj: is rand safe? */
static float
RandomizedSelect(float *A, int p, int r, int i)
{
    int k, q, m, n;
    float temp, x;

    /* Page 187 */
    if (p == r)
        return(A[p]);

    /* Page 162 */
    q = rand() % (r - p + 1) + p; 
    temp = A[p];
    A[p] = A[q];
    A[q] = temp;

    /* Page 154 */
    x = A[p];
    m = p;
    n = r;
    while (1) {
        while (A[n] > x)
            n--;
        while (A[m] < x)
            m++;
        if (m < n) {
            temp = A[m];
            A[m] = A[n];
            A[n] = temp;
            n--;
            m++;
        } else
            break;
    }

    /* Page 187 */
    k = n - p + 1;
    if (i <= k)
        return(RandomizedSelect(A,p,n,i));
    else
        return(RandomizedSelect(A,n+1,r,i-k));
}

/* This is a C implementation of the MATLAB file MakeQtrMask2.m.
   There is a bug in it somewhere. I rewrote it in Python.
   static double* MakeQtrMask(double r, int nwedges)
*/
                    
/**************************************************/
/***     Level 1 Functions                      ***/
/**************************************************/

/* This routine creates the mask that is one-quarter of the size of the circle
 * and nwedges deep.  It also creates a mask sum over all the wedges and
 * computes the number of pixels with non-zero weights.
 * Note that radius == (int) ceil(3 * sigma) always holds. Passing them
 * separately seems to be an quirk of history.
   Now written in Python.
static int
CreateMask(double sigma, int wedges, int radius, PyArrayObject** maskarr,
          PyArrayObject** sumarr)
*/
static int
ClusterPoints(PyArrayObject* Lab, PyArrayObject* masksum, 
              int radius, int totalpoints, int row, int col,
              double totalweight, float maxweight, int maxclusters, 
              PyArrayObject* cluster)
{
    float *weights, *temp;
    PyArrayObject* points;
    int i, j, nclusters, npoints, simple;
    int *index = NULL; 
    float ratio;
    int dims[3];
    
    dims[0] = totalpoints;
    dims[1] = DIM;
    points = (PyArrayObject*) PyArray_FromDims(2, dims, PyArray_FLOAT);
    if (!points) {
        PyErr_NoMemory();
        return -1;
    }

    weights = (float*) calloc(totalpoints, sizeof(float));
    if (!weights) {
        PyErr_NoMemory();
        Py_DECREF(points);
        return -1;
    }

    /* To get the proper sampling percentage, we have two options.  The
     * simple one is to multiply all the mask weights (which we treat as
     * probabilities) by a constant.  However, if this constant is too high,
     * the maximum weight will go above 1.0.  Therefore, we must use a more
     * complex option.
     */
    if (maxweight * totalpoints * SAMPLE_PERCENTAGE / totalweight > 1.0) {
        simple = 0;
        ratio = (1.0 - SAMPLE_PERCENTAGE) / (1.0 - totalweight / totalpoints);
    } else {
        simple = 1;
        ratio = (float) totalpoints * SAMPLE_PERCENTAGE / totalweight;
    }
    
    /* Determine if the window is large enough for sampling, and if so,
     * sample each pixel in the mask with a Bernoulli trial.  Otherwise,
     * use all the pixels for clustering.
     */
    /* Here we plop the mask over the image as a sliding window.  The window
       has even sides of size 2*radius. The four center pixels correspond to
       (row-1, col), (row-1, col-1), (row, col-1), (row, col). Therefore
       i = j = 0 will correspond to r-radius, c-radius. */
    npoints = 0;
    for (i = 0; i < 2*radius; i++)
        for (j = 0; j < 2*radius; j++)
            if (masksumSUB(i, j) > 0) {
                pointsSUB(npoints, 0) =
                     LabSUB(0, row-radius+j, col-radius+i);
                pointsSUB(npoints, 1) =
                     LabSUB(1, row-radius+j, col-radius+i);
                pointsSUB(npoints, 2) =
                     LabSUB(2, row-radius+j, col-radius+i);
                weights[npoints] =
                     masksumSUB(i, j);
                npoints++;
            }
    if (totalpoints != npoints) {
        printf("L258 totalpoints != npoints-1 %d %d\n", totalpoints, npoints);
        exit(0);
    }

    if (PRINT_SAMPLING)
        printf("Simple=%d Ratio: %.3f  Points: %d of %d  Pixel percentage: %.2f\n",
            simple, ratio, npoints, totalpoints,
            (double) npoints / (double) totalpoints);

    /* Cluster using binary split algorithm */
    /* Changed 18 May 99 to include indexing */
    /* Because index == NULL, bs creates array index internally. */
    bs((float*) points->data, npoints, maxclusters, &nclusters, &temp, index);

    /* Copy output from temp to cluster */
    for (i = 0; i < nclusters; i++)
        for (j = 0; j < DIM; j++) {
            clusterSUB(i, j) = temp[i * DIM + j];
        }

    Py_DECREF(points);
    free(weights);
    free(temp);

    return(nclusters);
}


static void
CreateWedgeHistograms(PyArrayObject* Lab, PyArrayObject* mask, int row, 
                      int col, int radius, int nwedges,
                      PyArrayObject* cluster, int nclusters,
                      PyArrayObject* hist)
{
    int i, j, k;
    float com[MAXCLUSTERS][DIM], sum[MAXCLUSTERS];

    memset(hist->data, 0, hist->dimensions[0] * hist->dimensions[1] * sizeof(double));
    memset(com, 0, MAXCLUSTERS * DIM * sizeof(float));
    memset(sum, 0, MAXCLUSTERS * sizeof(float));

    /* Here we plop the mask over the image as a sliding window.  The window
       has even sides of size 2*radius. The four center pixels correspond to
       (row-1, col), (row-1, col-1), (row, col-1), (row, col). Therefore
       i = j = 0 will correspond to r-radius, c-radius. */
    for (k = 0; k < 4*nwedges; k++) {
        for (i = 0; i < 2*radius; i++)
            for (j = 0; j < 2*radius; j++) {
                double value = maskSUB(k, i, j);
                if (value > 0) {
                    AddWeight(Lab, value, row-radius+j, col-radius+i,
                         cluster, hist, k, com, sum, nclusters);
                }
            }
    }
}

static void
ComputeCostMatrix(PyArrayObject* cluster, int nclusters)
{
    int i, j;
    float dist, temp_dist;

    /* cost has to be a global variable because of the way the EMD code is */
    memset(cost, 0, MAXCLUSTERS*MAXCLUSTERS*sizeof(double));

    for (i = 0; i < nclusters; i++) {
        for (j = 0; j < nclusters; j++) {
            if (i > j)
                cost[i][j] = cost[j][i];
            else if (i == j || (clusterSUB(i, 0) == BLACK &&
                    clusterSUB(j, 0) == BLACK))
                cost[i][j] = 0.0;
            else {
                dist = 0;
                temp_dist = (clusterSUB(i, 0) -
                        clusterSUB(j, 0));
                dist += temp_dist * temp_dist; 
                temp_dist = (clusterSUB(i, 1) -
                        clusterSUB(j, 1));
                dist += temp_dist * temp_dist;
                temp_dist = (clusterSUB(i, 2) -
                        clusterSUB(j, 2));
                dist += temp_dist * temp_dist;

                /* dist = sqrt(dist) / PERCEPTUAL_THRESH; 
                      cost[i][j] = (dist >= 1) ? 1 : dist;*/

                cost[i][j] = 1 - exp(-sqrt(dist) / GAMMA);
                /* cost[i][j] = sqrt(dist) / 100; */
            }
            if (PRINT_COST)
                printf("%.2f ",cost[i][j]);
            }
        if (PRINT_COST)
            printf("\n");
    }
    if (PRINT_COST)
        printf("\n");
}

static void
ComputeOutputParameters(float *work, int nwedges, int nori,
                        int row, int col,
                        PyArrayObject* str, PyArrayObject* ab,
                        PyArrayObject* unc, PyArrayObject* ori)
{
    int i, strindex = 0, abindex = 0, l_uncert, r_uncert, responses;
    int startindex = 0;
    double angle, maxEMD = 0.0, minEMD = 1.0, unc_thresh, maxangle;
    double a, b, c, strength, orientation, dangle, unc_length;
    double wedgesize, maxEMDori = -1.0; /* Quiets gcc */

    wedgesize = 90.0 / nwedges;
    maxangle = (double)(nori * wedgesize); /* Always 180 or 360 */

    /* Compute Minimum and Maximum EMD values */
    for (i = 0; i < nori; i++) {
        if (work[i] > maxEMD) {
            maxEMD = work[i];
            strindex = i;
            maxEMDori = strindex * wedgesize;
        }
        if (work[i] < minEMD) {
            minEMD = work[i];
            abindex = i;
        }
    }

    /* The strength and orientation of an edge (or a corner) lie not at the
     * maximum EMD value but rather at the vertex of the parabola that runs
     * through the maximum and the two points on either side.  The first
     * computation of orientation assumes the maximum is the y-intercept.  
     * After computing the strength, we adjust the orientation.  
     */
    a = work[strindex];
    b = work[(strindex+nori-1) % nori];
    c = work[(strindex+nori+1) % nori];
    if (b + c != 2 * a) {
        orientation = (wedgesize / 2) * ((b - c) / (b + c - 2 * a));
        strength = a + ((c - b) / (2 * wedgesize)) * orientation +
            ((b + c - 2 * a) / (2 * wedgesize * wedgesize)) * orientation * 
              orientation;
        orientation = fmod(maxEMDori + orientation + 360, maxangle);
    } else { /* Uncertainty abounds */
        strength = a;
        orientation = maxEMDori;
    }
    strSUB(row, col, 0) = strength;
    if (ab)
        abSUB(row, col, 0) = minEMD;

    /* Compute Uncertainties and Orientations */
    unc_thresh = 0.95 * maxEMD;

    /* First, find a value below the threshold, or declare total uncertainty */
    startindex = 0;
    while (startindex < nori && work[startindex] > unc_thresh)
        startindex++;
    if (startindex == nori) { /* Total uncertainty */
        if (unc)
            uncSUB(row, col, 0) = maxangle;
        if (ori)
            for (responses = 0; responses < MAXRESPONSES; responses++)
                oriSUB(row, col, responses) = -1;
        return;
    }

    /* Now scan until we find values above the threshold and record them */
    /* Do this as many times as needed */
    responses = 0;
    i = startindex;
    while (responses < MAXRESPONSES) {
        while (work[(i + nori) % nori] <= unc_thresh && i != startindex + nori)
            i++;
        l_uncert = i;
        if (i == startindex + nori) /* Finished all orientations */
            break;
        while (work[(i + nori) % nori] > unc_thresh && i != startindex + nori)
            i++;
        r_uncert = i - 1;

        /* Uncertainty exists only if three or more consecutive values are
         * all above unc_thresh (meaning r_uncert - l_uncert is 2 or
         * greater).  Otherwise, the strength and orientation estimation
         * routine above can handle it.
         */
        if (unc) {
            if (r_uncert - l_uncert > 1) {
                uncSUB(row, col, responses) = (r_uncert - l_uncert) * wedgesize;
                if (uncSUB(row, col, responses) > maxangle) {
                    uncSUB(row, col, responses) = maxangle;
                }
            } else {
                uncSUB(row, col, responses) = 0;
            }
        }

        /* Compute Orientation 
         * Our orientation estimate can only be used when we are near the
         * maximum; other high responses must use a more roughly quantized
         * value.
         */
        if (ori) {
            angle = (r_uncert + l_uncert) * wedgesize / 2;
            unc_length = r_uncert - l_uncert;
            while (angle < 0)
                angle += maxangle;
            while (angle >= maxangle)
                angle -= maxangle;
            dangle = MIN(abs(maxEMDori - angle), 
                               abs(maxEMDori - (angle + maxangle)));
            if (unc_length < 2 && dangle <= unc_length * wedgesize / 2 ) { 
                /* Use orientation estimate */
                oriSUB(row, col, responses) = orientation;
            } else {
                oriSUB(row, col, responses) = angle;
            }
        }
        responses++;
    }

    if (ori) /* Fill the remaining values with -1's */
        while (responses < MAXRESPONSES) {
            oriSUB(row, col, responses) = -1;
            responses++;
        }
}

/* "*mask" and "*summarr" were gotten using PySequence_GetItem. Therefore
   Py_DECREF "*mask" and "*summarr" when you are done with them. */
static void
ExtractMaskData(PyObject* maskdata, int n,
     PyArrayObject** mask, PyArrayObject** sumarr, int* npoints)
{
    PyObject* item;
    PyObject* temp;

    item = PySequence_GetItem(maskdata, 0);
    temp = PySequence_GetItem(item, n);
    *npoints = PyInt_AsLong(temp);
    Py_DECREF(item);

    item = PySequence_GetItem(maskdata, 1);
    (*mask) = (PyArrayObject*) PySequence_GetItem(item, n);
    Py_DECREF(item);

    item = PySequence_GetItem(maskdata, 2);
    (*sumarr) = (PyArrayObject*) PySequence_GetItem(item, n);
    Py_DECREF(item);
}

/* numsigmas >= 1
   len(sigmas) == numsigmas
   len(spacing) == numsigmas
   spacing defaults to 1
   len(angles) = numangles  default is 1
   angles[0] defaults to 180
   0 < angles[i] <= 180
   default nwedges == 6
   len(maxclusters) == numsigmas
*/
static void
_Compass(PyObject* arr, PyObject* maskdata,
         double *sigmas, int numsigmas, int maxradius, double *spacing, 
         double *angles, int numangles, int nwedges, 
         double *maxclusters, PyObject* pyout[4]) 
{
    PyArrayObject* Lab = (PyArrayObject*) arr;
    PyArrayObject* strength = (PyArrayObject*) pyout[0];
    PyArrayObject* orientation = (PyArrayObject*) pyout[1];
    PyArrayObject* abnormality = (PyArrayObject*) pyout[2];
    PyArrayObject* uncertainty = (PyArrayObject*) pyout[3];
    int rows, cols, npoints, outputrows, outputcols, mc;
    int h, i, j, row, rr, col, cc, s, space, nclusters, nori, anglewedges, masksz;
    int dims[3];
    PyArrayObject* hist;
    PyArrayObject* mask;
    PyArrayObject* masksum;
    PyArrayObject* str = NULL;
    PyArrayObject* ab = NULL;
    PyArrayObject* unc = NULL;
    PyArrayObject* ori = NULL;
    double sum, sigma;
    PyArrayObject* cluster;
    float hist1[MAXCLUSTERS], hist2[MAXCLUSTERS];
    float hist1norm[MAXCLUSTERS], hist2norm[MAXCLUSTERS];
    float *work, maxwt = 0.0, inmass, outmass;
    feature_t f[MAXCLUSTERS];
    signature_t s1, s2;

    if (Lab->descr->type_num != PyArray_FLOAT) {
        PyErr_SetString(PyExc_TypeError, "input array must be a PyArray_FLOAT");
        return;
    }
    if (Lab->nd != 3) {
        PyErr_SetString(PyExc_TypeError, "input array must have 3 dims.");
        return;
    }
    if (Lab->dimensions[0] != 3) {
        PyErr_SetString(PyExc_TypeError, "array.shape[0] must be 3.");
        return;
    }

    rows = Lab->dimensions[1];
    cols = Lab->dimensions[2];

    /* 4 * nwedges is the maximum number of orientations */
    work = (float*) calloc(4 * nwedges, sizeof(float));
    if (!work) {
        PyErr_NoMemory();
        return;
    }

    dims[0] = 4*nwedges;
    dims[1] = MAXCLUSTERS;
    hist = (PyArrayObject*) PyArray_FromDims(2, dims, PyArray_DOUBLE);
    if (!hist) {
        PyErr_NoMemory();
        free(work);
        return;
    }

    for (i = 0; i < MAXCLUSTERS; i++)
        f[i] = i;
    s1.Features = f;
    s2.Features = f;
    s1.Weights = hist1norm;
    s2.Weights = hist2norm;
    srand(clock());

    dims[0] = MAXCLUSTERS;
    dims[1] = DIM;
    cluster = (PyArrayObject*) PyArray_FromDims(2, dims, PyArray_FLOAT);
    if (cluster == NULL) {
        Py_DECREF(hist);
        free(work);
        return;
    }

    /* Loop over every sigma value */
    for (s = 0; s < numsigmas; s++) {
        /* Retrieve data for this sigma */
        /* Rewritten 31 July 1999 to allow floating-point values for sigma */
        sigma = sigmas[s];
        masksz = (int) ceil(3 * sigma);  
        space = (int) spacing[s];
        mc = (int) maxclusters[s];

        ExtractMaskData(maskdata, s, &mask, &masksum,  &npoints);

        /* Compute max weight of entire circle and sum of weights over one
         * wedge of the circle.
         */
        sum = 0.0;
        for (i=0; i<masksz; i++) /* sums over 1/4 of circle */
            for (j=0; j<masksz; j++) {
                sum += masksumSUB(i, j);
                if (masksumSUB(i, j) > maxwt)
                    maxwt = masksumSUB(i, j);
        }
        sum /= nwedges;

        /* Loop over every desired image position */
        for (row = maxradius; row <= rows - maxradius; row += space) {
            if ((row+1) % 50 == 0)
                printf("row %d\n", row+1);
            for (col = maxradius; col <= cols - maxradius; col += space) {
                cc = col;
                if (col > maxradius) cc = col - col%2;
                rr = maxradius;
                nclusters = ClusterPoints(Lab, masksum, masksz,
                            npoints, row, col, sum * nwedges * 4,
                            maxwt, mc, cluster);

                CreateWedgeHistograms(Lab, mask, row, col, masksz,
                          nwedges, cluster, nclusters, hist);

                ComputeCostMatrix(cluster, nclusters);

                /* Loop over each desired angle */
                for (h = 0; h < numangles; h++) {
                    /* Compute desired data and cell pointers for each angle */
                    /* "anglewedges" is "angles per wedge". */
                    anglewedges = angles[h] * nwedges / 90.0;
                    inmass = sum * ((anglewedges <= 2) ? anglewedges * WEDGEWT :
                                         WEDGEWT * 2 + anglewedges - 2);
                    outmass = sum * (WEDGEWT * 2 + 4 * nwedges - anglewedges - 2); 
                    nori = (angles[h] == 180) ? 2 * nwedges : 4 * nwedges;
                    str = (PyArrayObject*)
                          strengthSUB(s, h);
                    outputrows = str->dimensions[0];
                    outputcols = str->dimensions[1];
                    if (abnormality) {
                        ab = (PyArrayObject*)
                              abnormalitySUB(s, h);
                    }
                    if (orientation) {
                        ori = (PyArrayObject*)
                              orientationSUB(s, h);
                    }
                    if (uncertainty) {
                        unc = (PyArrayObject*)
                              uncertaintySUB(s, h);
                    }
                    /* Compute initial histogram sums */
                    ckindex("L676", nclusters, MAXCLUSTERS);
                    for (i = 0; i < nclusters; i++) {
                        hist1[i] = 0.0;
                        for (j = 0; j < anglewedges; j++)
                            hist1[i] += (float) histSUB(j, i);

                        hist2[i] = 0.0;
                        for (j = anglewedges; j < 4 * nwedges; j++)
                            hist2[i] += (float) histSUB(j, i);
                    }

                    /* Loop over every orientation */
                    ckindex("L690", nclusters, MAXCLUSTERS);
                    for (i = 0; i < nori; i++) {
                        /* Normalize the histograms */
                        for (j = 0; j < nclusters; j++) {
                            hist1norm[j] = hist1[j] / inmass;
                            if (PARTIAL_MATCH) /* Normalize both by same amount */
                                hist2norm[j] = hist2[j] / inmass;
                            else  /* Normalize both by the number of wedges */
                                hist2norm[j] = hist2[j] / outmass;
                        }

                        if (PRINT_HIST) {
                            printf("Hist\n");
                            for (j = 0; j < nclusters; j++)
                                printf("%.2f ", hist1norm[j]);
                            printf("\n");
                            for (j = 0; j < nclusters; j++)
                                printf("%.2f ", hist2norm[j]);
                            printf("\n");
                        }
                    
                        /* Compute EMD */
                        s1.n = nclusters;
                        s2.n = nclusters;
                        ckindex("L723", i, 4*nwedges);
                        work[i] = emd(&s1, &s2, dist, NULL, NULL);
                        if (work[i] > 1)
                            work[i] = 1.0;
                        else if (work[i] < 0)
                            work[i] = 0.0;
                    
                        /* Update the histograms except for the last iteration */
                        if (i < nori - 1)
                            for (j = 0; j < MAXCLUSTERS; j++) {
                                hist1[j] += -histSUB(i, j)
                                  + histSUB((i+anglewedges)%(4*nwedges), j);
                                hist2[j] += -histSUB((i+anglewedges)%(4*nwedges), j)
                                  + histSUB(i, j);
                            }
                    }
                    ComputeOutputParameters(work, nwedges, nori, 
                        (row - maxradius) / space, (col - maxradius) / space,
                        str, ab, unc, ori);
                    /* The original code was:
                       ComputeOutputParameters(work, nwedges, nori, 
                          (c - maxradius) / space * outputrows +
                               (r - maxradius) / space, 
                          outputrows * outputcols, plot,
                          str, ab, unc, ori);
                       MatLAB uses (row, col) notation but FORTRAN array
                       ordering. Since I access everything by pixels, I
                       don't care. I hope I am right.
                    */
                    /* For debugging, NULL these pointers I am temporarily
                       finished with. */
                    str = NULL;
                    ab = NULL;
                    unc = NULL;
                    ori = NULL;
                } /* Next angle */
            } /* Next image position */
        }
        Py_DECREF((PyObject*) mask);
        Py_DECREF((PyObject*) masksum);
    } /* Next sigma */

    free(work);
    Py_DECREF((PyObject*) hist);
    Py_DECREF((PyObject*) cluster);
}
%}

/* See defines.i for this SWIG multi-line define. */
LISTIN(double, numsigmas, sigmas, PyFloat_AsDouble)
LISTIN(double, numspacing, spacing, PyFloat_AsDouble)
LISTIN(double, numangles, angles, PyFloat_AsDouble)
LISTIN(double, nummaxclusters, maxclusters, PyFloat_AsDouble)

%inline %{
/* arrin->dimension[0] == 3
   1 <= num_outputs < 4
   numsigmas >= 1
   len(sigmas) == numsigmas
   len(spacing) == numsigmas
   spacing defaults to 1
   len(angles) = numangles  default is 1
   angles[0] defaults to 180
   0 < angles[i] <= 180
   default nwedges == 6
   len(maxclusters) == numsigmas
*/
PyObject* Compass(PyObject* arrin, PyObject* maskdata,
         int num_outputs, int nwedges,
         int numsigmas, double *sigmas,
         int numspacing, double *spacing, 
         int numangles, double *angles,
         int nummaxclusters, double *maxclusters)
{
    PyObject** pyout;
    PyObject* arr;
    PyObject* pytuple;
    int rows, cols, i, maxradius;
    double maxsigma = -1.0;

    /* numspacing and nummaxclusters exist only to make SWIG wrapping
       easier.
    */
    if (numsigmas != numspacing || numsigmas != nummaxclusters) {
        PyErr_SetString(PyExc_TypeError, "sigmsa, spacing, and "
          "maxclusters must be the same length.");
        return NULL;
    }

    for (i = 0; i < numsigmas; i++)
        if (sigmas[i] > maxsigma)
            maxsigma = sigmas[i];
    maxradius = ceil(3 * maxsigma);

    arr = ConvertImage_RGB2Lab(arrin);
    if (arr == NULL) return NULL;

    rows = ((PyArrayObject*) arr)->dimensions[1];
    cols = ((PyArrayObject*) arr)->dimensions[2];
    pyout = CreateOutputArrays(numsigmas, numangles, rows, cols,
           num_outputs, spacing, maxradius, nwedges);
    _Compass(arr, maskdata,
           sigmas, numsigmas, maxradius, spacing, 
           angles, numangles, nwedges, 
           maxclusters, pyout);
    if (PyErr_Occurred()) return NULL;

    pytuple = PyTuple_New(num_outputs);
    if (pytuple == NULL) {
        for (i=0; i<num_outputs; i++)
            Py_DECREF(pyout[i]);
        free(pyout);
        return NULL;
    }
    for (i=0; i<num_outputs; i++)
        PyTuple_SET_ITEM(pytuple, i, pyout[i]);
    free(pyout);
    return pytuple;
}
%}

/* See defines.i for this SWIG multi-line define. */
LISTCLEAR(double, numsigmas, sigmas)
LISTCLEAR(double, numspacing, spacing)
LISTCLEAR(double, numangles, angles)
LISTCLEAR(double, nummaxclusters, maxclusters)
